"""Simple reader-writer locks in Python
Many readers can hold the lock XOR one and only one writer"""
import threading
 
version = """$Id: 04-1.html,v 1.3 2006/12/05 17:45:12 majid Exp $"""
 
class RWLock:
  """
A simple reader-writer lock Several readers can hold the lock
simultaneously, XOR one writer. Write locks have priority over reads to
prevent write starvation.
"""
  def __init__(self):
    self.rwlock = 0
    self.writers_waiting = 0
    self.monitor = threading.Lock()
    self.readers_ok = threading.Condition(self.monitor)
    self.writers_ok = threading.Condition(self.monitor)
  def acquire_read(self):
    """Acquire a read lock. Several threads can hold this typeof lock.
It is exclusive with write locks."""
    self.monitor.acquire()
    while self.rwlock < 0 or self.writers_waiting:
      self.readers_ok.wait()
    self.rwlock += 1
    self.monitor.release()
  def acquire_write(self):
    """Acquire a write lock. Only one thread can hold this lock, and
only when no read locks are also held."""
    self.monitor.acquire()
    while self.rwlock != 0:
      self.writers_waiting += 1
      self.writers_ok.wait()
      self.writers_waiting -= 1
    self.rwlock = -1
    self.monitor.release()
  def promote(self):
    """Promote an already-acquired read lock to a write lock
    WARNING: it is very easy to deadlock with this method"""
    self.monitor.acquire()
    self.rwlock -= 1
    while self.rwlock != 0:
      self.writers_waiting += 1
      self.writers_ok.wait()
      self.writers_waiting -= 1
    self.rwlock = -1
    self.monitor.release()
  def demote(self):
    """Demote an already-acquired write lock to a read lock"""
    self.monitor.acquire()
    self.rwlock = 1
    self.readers_ok.notifyAll()
    self.monitor.release()
  def release(self):
    """Release a lock, whether read or write."""
    self.monitor.acquire()
    if self.rwlock < 0:
      self.rwlock = 0
    else:
      self.rwlock -= 1
    wake_writers = self.writers_waiting and self.rwlock == 0
    wake_readers = self.writers_waiting == 0
    self.monitor.release()
    if wake_writers:
      self.writers_ok.acquire()
      self.writers_ok.notify()
      self.writers_ok.release()
    elif wake_readers:
      self.readers_ok.acquire()
      self.readers_ok.notifyAll()
      self.readers_ok.release()



# encoding: utf-8

import inspect
import logging
import pkg_resources
import socket

__all__ = [
    'EndPoint',
    'PPrintEndPoint'
]

class EndPoint(object):
    """Abstract class meant to receive containers metrics
    sent by `collector.ContainerStatsEmitter`

    It loads metrics registered as plugins to enrich metrics provided by collector.

    You need to implement the `emit` member method.
    """
    def __init__(self):
        self._logger = logging.getLogger("end-point")
        self.metrics_plugins = self._load_metrics_plugins()

    IGNORED_METRIC_KEYS = {'name', 'timestamp', 'stats'}
    METRICS_GROUP = 'docker_zabbix_sender.metrics'
    EVENT_KEY_PREFIX = 'docker.container.'
    HOST_FQDN = socket.getfqdn()

    @classmethod
    def container_hostname(cls, container_name):
        """Get "real" hostname of a container.

        Default implement returns: 'docker-{container_name}.{current_host_fqdn}'

        :param: metrics of a container given to the endpoint (dict)
        """
        return "{0}.docker.{1}".format(container_name, cls.HOST_FQDN)

    def __call__(self, client, containers_metrics):
        """Method used by collector to emit new metrics to this end-point.
        Metrics are translated, enriched, and passed to the `emit` member method

        :param client: Docker client

        :params containers_metrics: list of dict with containers information, one dict per container.
        """
        events, statistics = self._metrics_to_events(containers_metrics)
        self._enrich_with_plugins(client, statistics, events)
        self.emit(events)

    def emit(self, events):
        """
        :param events: list of dict with the following keys: hostname, timestamp, key, value
        """
        raise NotImplementedError()

    def close(self):
        """Release allocated resources. Meant to be overloaded
        """
        pass

    def _metrics_to_events(self, containers_metrics):
        """Transform list of dict containing containers metrics to a list of dict with the following keys:
            'host', 'key', 'value'
        Each dict also contains the 'timestamp' key if the metrics provide it.

        :param containers_metrics: new metrics given to the endpoint
        :return list of events
        """
        events = []
        stats = []
        for metrics in containers_metrics:
            stats.append(metrics['stats'])
            hostname = EndPoint.container_hostname(metrics['name'])
            timestamp = metrics['timestamp']
            for key, value in metrics.items():
                if key in EndPoint.IGNORED_METRIC_KEYS:
                    continue
                events.append({
                    'hostname': hostname,
                    'timestamp': timestamp,
                    'key': EndPoint.EVENT_KEY_PREFIX + key,
                    'value': value,
                })
        return events, stats

    def _enrich_with_plugins(self, client, statistics, events):
        """Ask registered metrics plugins to produce additional events according to new containers metrics

        :param client: Docker client given to the metrics plugins

        :param statistics: list of tuple providing container statistics, one dict per container.

        :param events: events collection, to be filled by metrics plugins
        """
        for name, collector in self.metrics_plugins.items():
            try:
                events.extend(collector(EndPoint.HOST_FQDN, client, statistics))
            except Exception, e:
                self._logger.exception("Could not collect metrics from plugin %s", name)

    def _load_metrics_plugins(self):
        """Loads objects registered with the '[docker-zabbix-sender.metrics]' entry point.
        :return dict of plugin_name -> callable_object
        """
        metrics = {}
        for entrypoint in pkg_resources.iter_entry_points(group=EndPoint.METRICS_GROUP):
            try:
                plugin = entrypoint.load()
                metrics[entrypoint.name] = plugin
                assert hasattr(plugin, '__call__')
                self._logger.info("Registered metrics entrypoint '%s'", entrypoint.name)
            except Exception:
                self._logger.exception("Could not load entrypoint %s", entrypoint.name)
        return metrics


class PPrintEndPoint(EndPoint):
    """Dumb EndPoint that prints produced events"""
    def emit(self, events):
        import pprint
        pprint.pprint(events)



# encoding: utf-8

import json
import logging
import time
import threading

from docker import Client

from .RWLock import RWLock

__all__ = [
    'ContainerStatsEmitter',
    'ContainerStats'
]

class ContainerStats(threading.Thread):
    """Provides a set of metrics about a Docker container.

    Those metrics are updated repeatedly (about every second) by
    `stats` method available in Docker remote API since v17.

    TODO: make it a thread
    """

    def __init__(self, container, docker):
        """
        :param container: The Docker container identifier to monitor.

        :param docker: Docker client
        :type docker: Client
        """
        threading.Thread.__init__(self)
        self.container = container
        self.name = docker.inspect_container(container)['Name'][1:]
        self.user_cpu_percent = 0.0
        self.kernel_cpu_percent = 0.0
        self.memory = 0.0
        self.memory_limit = 0.0
        self.memory_percent = 0.0
        self.network_rx = 0.0
        self.network_tx = 0.0
        self.stats = None
        self.timestamp = int(time.time())
        self._docker = docker
        self._lock = RWLock()
        self._response = None

    def run(self):
        """Collect container metrics repeatedly. Does not returns
        unless the Docker stats stream is closed or the `shutdown` method
        is called.
        """
        previous_user_cpu = 0.0
        previous_kernel_cpu = 0.0
        previous_total_cpu = 0.0
        previous_system = 0.0
        previous_network_rx = 0.0
        previous_network_tx = 0.0
        start = True
        url = self._docker._url("/containers/{0}/stats".format(self.container))
        self._response = self._docker._get(url, stream=True)
        stream = self._docker._stream_helper(self._response, decode=True)
        try:
            for stats in stream:
                stats['timestamp']= int(time.time())
                # Provides additional fields that can be used by metrics plugins
                stats['name'] = self.name
                stats['id'] = self.container
                # code below is strongly inspired from docker's code.
                mem_percent = float(stats['memory_stats']['usage']) / stats['memory_stats']['limit'] * 100.0
                user_cpu_percent = 0.0
                kernel_cpu_percent = 0.0
                if not start:
                    user_cpu_percent, kernel_cpu_percent = self._calculate_cpu_percent(
                        previous_user_cpu,
                        previous_kernel_cpu,
                        previous_system,
                        stats
                    )
                start = False
                self._lock.acquire_write()
                self.timestamp = stats['timestamp']
                self.stats = stats
                self.user_cpu_percent = user_cpu_percent
                self.kernel_cpu_percent = kernel_cpu_percent
                self.memory = float(stats['memory_stats']['usage'])
                self.memory_limit = float(stats['memory_stats']['limit'])
                self.memory_percent = mem_percent
                self.network_rx = float(stats['network']['rx_bytes']) - previous_network_rx
                self.network_tx = float(stats['network']['tx_bytes']) - previous_network_tx
                self._lock.release()
                previous_user_cpu = stats['cpu_stats']['cpu_usage']['usage_in_usermode']
                previous_kernel_cpu = stats['cpu_stats']['cpu_usage']['usage_in_kernelmode']
                previous_total_cpu = stats['cpu_stats']['cpu_usage']['total_usage']
                previous_system = stats['cpu_stats']['system_cpu_usage']
                previous_network_rx = float(stats['network']['rx_bytes'])
                previous_network_tx = float(stats['network']['tx_bytes'])
        except AttributeError:
            # raise in urllib3 when the stream is closed while waiting for stuff to read
            pass
        finally:
            self.shutdown() # ensure stream is closed

    def emit(self, consumer_func):
        """Provide consumer access to the container stats.

        :param: `consumer_func`: callable instance. This parameter is called with
        this instance given in parameter.

        TODO: pass a dict() of values directly.
        """
        self._lock.acquire_read()
        try:
            consumer_func(self)
        finally:
            self._lock.release()


    def shutdown(self):
        """Stop collecting the container metrics.
        """
        if self._response:
            self._response.raw.close()
            self._response = None

    def _calculate_cpu_percent(self,
        previous_user_cpu,
        previous_kernel_cpu,
        previous_system,
        stats
    ):
        """Compute CPU percentage.

        strongly inspired from docker's code.

        :return: cpu percentage
        """
        user_cpu_percent = 0.0
        kernel_cpu_percent = 0.0
        # calculate the change for the cpu usage of the container in between readings
        user_cpu_delta = float(stats['cpu_stats']['cpu_usage']['usage_in_usermode']) - previous_user_cpu
        kernel_cpu_delta = float(stats['cpu_stats']['cpu_usage']['usage_in_kernelmode']) - previous_kernel_cpu
        # calculate the change for the entire system between readings
        system_delta = float(stats['cpu_stats']['system_cpu_usage']) - previous_system
        if system_delta > 0.0:
            if user_cpu_delta > 0.0:
                user_cpu_percent = (user_cpu_delta / system_delta) * float(len(stats['cpu_stats']['cpu_usage']['percpu_usage'])) * 100.0
            if kernel_cpu_delta > 0.0:
                kernel_cpu_percent = (kernel_cpu_delta / system_delta) * float(len(stats['cpu_stats']['cpu_usage']['percpu_usage'])) * 100.0
        return user_cpu_percent, kernel_cpu_percent


class ContainerStatsEmitter(threading.Thread):
    """Maintain a list of `ContainerStats` collecting metrics of several Docker containers.
    Repeatedly aggregates all metrics and push them to a consumer. The list of container collectors
    is updated according to containers started, stopped, ...
    """

    def __init__(self, client, endpoint_func, delay=30):
        """
        :param client: Docker client

        :param endpoint_func: a callable instance which is given aggregated statistics repeadetly.
        The endpoint_func may also have a 'close' callable attribute.

        :param delay: Number of seconds between 2 notifications of `endpoint_func`
        """
        threading.Thread.__init__(self)
        self._client = client
        self._endpoint_func = endpoint_func
        self._delay = delay
        self._stop = False
        self._logger = logging.getLogger("stats-emitter")

    def run(self):
        container_stats = dict()
        try:
            while self._should_run():
                # update list of container stats
                running_containers = set(map(lambda c: c['Id'], self._client.containers()))
                monitored_containers = set(container_stats.keys())
                started_containers = running_containers - monitored_containers
                stopped_containers = monitored_containers - running_containers
                for container in stopped_containers:
                    self._logger.info("container has stopped: %s", container)
                    container_stats.pop(container).shutdown()
                for container in started_containers:
                    self._logger.info("Monitoring activity of container: %s", container)
                    stats = ContainerStats(container, self._client)
                    container_stats[container] = stats
                    stats.start()
                time.sleep(self._delay)
                if not self._should_run():
                    return
                # collect results
                payload = []
                def append(stats):
                    payload.append({
                        'name': stats.name,
                        'id': stats.container,
                        'stats': stats.stats,
                        'cpu.user_percent': stats.user_cpu_percent,
                        'cpu.kernel_percent': stats.kernel_cpu_percent,
                        'memory.used': stats.memory,
                        'memory.limit': stats.memory_limit,
                        'memory.percent': stats.memory_percent,
                        'network_rx': stats.network_rx,
                        'network_tx': stats.network_tx,
                        'timestamp': stats.timestamp,
                    })
                for stats in container_stats.values():
                    stats.emit(append)
                # emit to endpoint_func
                self._endpoint_func(self._client, payload)
        finally:
            self._logger.info("waiting for all collectors threads to terminate.")
            for container in container_stats.values():
                container.shutdown()
            for container in container_stats.values():
                container.join()
            if hasattr(self._endpoint_func, 'close'):
                self._endpoint_func.close()
            self._logger.info("collectors terminated successfully. See you bye!")

    def shutdown(self):
        """Ask thread termination. Method returns immediatly. You may
        call the `Thread.join` method afterward."""
        self._logger.info("user asked for daemon termination.")
        self._stop = True

    def _should_run(self):
        """Internal method used to know if the show must go on"""
        return not self._stop



# encoding: utf-8

"""Provides collection of events emitters"""

import time
from . import EndPoint

def containers_count(host_fqdn, docker_client, statistics):
    """
    Emit events providing:
      - number of containers
      - number of running containers
      - number of crashed containers

    :param host_fqdn: FQDN of the host where the docker-zabbix-daemon is running, for instance docker.acme.com
    :type host_fqdn: string

    :param docker_client: instance of docker.Client see http://docker-py.readthedocs.org/en/latest/api/
    :type docker_client: docker.Client

    :param statistics: List of dicts providing collected container statistics. see Docker stats API call on https://docs.docker.com/reference/api/docker_remote_api_v1.17/#get-container-stats-based-on-resource-usage

    :return: list of dicts providing additional events to push to Zabbix.
    Each dict is composed of 4 keys:
        - hostname
        - timestamp
        - key
        - value
    """
    running = 0
    crashed = 0
    now = int(time.time())
    containers = docker_client.containers(all=True)
    for container in containers:
        status = container['Status']
        if status.startswith('Up'):
            running += 1
        elif not status.startswith('Exited (0)'):
            crashed += 1
    data = {
        'all': len(containers),
        'running': running,
        'crashed': crashed,
    }
    return [
        {
            'hostname': '-',
            'timestamp': now,
            'key': EndPoint.EVENT_KEY_PREFIX + 'count.' + key,
            'value': value
        }
        for key, value in data.items()
    ]

def cpu_count(host_fqdn, docker_client, statistics):
    """Emit the number of CPU available for each container.
    """
    for stat in statistics:
        yield {
            'hostname': EndPoint.container_hostname(stat['name']),
            'timestamp': stat['timestamp'],
            'key': EndPoint.EVENT_KEY_PREFIX + 'cpu.count',
            'value': len(stat['cpu_stats']['cpu_usage']['percpu_usage']),
        }



# encoding: utf-8

import argparse
import logging
import os
import subprocess
import signal
import sys
import tempfile

from docker import Client
from docker.utils import kwargs_from_env

from .endpoint import EndPoint
from .collector import ContainerStatsEmitter

LOGGER = logging.getLogger(__name__)

class ZabbixSenderProcess(subprocess.Popen):
    """A 'zabbix_sender' process with stdin piped.
    """
    def __init__(self, **kwargs):
        subprocess.Popen.__init__(self, ZabbixSenderProcess.cmdline(**kwargs), stdin=subprocess.PIPE)

    @classmethod
    def cmdline(cls,
            input_file,
            config_file=None,
            zabbix_server=None,
            port=None,
            with_timestamps=False,
            real_time=False):
        """
        :return collection containing the 'zabbix_sender' command line to execute.
        """
        cmdline = ['zabbix_sender', '--input-file', input_file]
        if config_file:
            cmdline.extend(['--config', config_file])
        if zabbix_server:
            cmdline.extend(['--zabbix-server', zabbix_server])
        if port:
            cmdline.extend(['--port', port])
        if with_timestamps:
            cmdline.append('--with-timestamps')
        if real_time:
            cmdline.append('--real-time')
        return cmdline


class ZabbixSenderEndPoint(EndPoint):
    def __init__(self, **kwargs):
        """
        :param kwargs: optional arguments given to the `zabbix_sender` function.
        """
        EndPoint.__init__(self)
        self.zabbix_sender_p = ZabbixSenderProcess(
            input_file='-',
            with_timestamps=True,
            **kwargs
        )

    def emit(self, events):
        if not any(events):
            return
        fmt = "{hostname} {key} {value}\n"
        if events[0].has_key('timestamp'):
            fmt = "{hostname} {key} {timestamp} {value}\n"
        for event in events:
            self.zabbix_sender_p.stdin.write(fmt.format(**event))

    def close(self):
        self.zabbix_sender_p.communicate()

def run(args=None):
    """Main entry point. Runs until SIGTERM or SIGINT is emitted.

    :param args: Optional arguments, use `sys.argv[1:]` otherwise
    """
    if args is None:
        args = sys.argv[1:]
        FORMAT = '%(asctime)-15s %(levelname)-8s %(name)s %(message)s'
        logging.basicConfig(format=FORMAT, level=logging.INFO)

    from .version import version
    parser = argparse.ArgumentParser(
        description="""Provides Zabbix Docker containers statistics running on a Docker daemon.""")
    parser.add_argument("--tlsverify",
        action='store',
        choices=["true", "false"],
        default='true',
        help="Use TLS and verify the remote Docker daemon. Default is %(default)s"
    )
    parser.add_argument('-V', '--version',
        action='version',
        version='%(prog)s ' + version
    )
    parser.add_argument('-c', '--config',
        metavar="<file>",
        help="Absolute path to the zabbix agent configuration file"
    )
    parser.add_argument('-z', '--zabbix-server',
        metavar='<server>',
        help='Hostname or IP address of Zabbix server'
    )
    parser.add_argument('-p', '--port',
        metavar='<server port>',
        help='Specify port number of server trapper running on the server. Default is 10051'
    )
    parser.add_argument('-i', '--interval',
        metavar='<sec>',
        default=30,
        type=int,
        help='Specify Zabbix update interval (in sec). Default is %(default)s'
    )
    parser.add_argument('-r', '--real-time',
        action='store_true',
        help="zabbix_sender push metrics to Zabbix one by one as soon as they are sent."
    )
    args = parser.parse_args(args)
    kwargs  = kwargs_from_env()
    if not args.tlsverify.lower() in ("yes", "true", "t", "1"):
        kwargs['tls'].assert_hostname = False
    docker_client = Client(**kwargs)
    docker_client.info()

    from zabbix_sender import ZabbixSenderEndPoint
    emitter = ContainerStatsEmitter(
        docker_client,
        ZabbixSenderEndPoint(
            config_file=args.config,
            zabbix_server=args.zabbix_server,
            port=args.port,
            real_time=args.real_time,
        ),
        args.interval)
    def _stop_emitter(signum, frame):
        """Handle for signal catching used to stop the `ContainerStatsEmitter` thread
        """
        emitter.shutdown()
    signal.signal(signal.SIGTERM, _stop_emitter)
    signal.signal(signal.SIGINT, _stop_emitter)
    emitter.start()
    signal.pause()

if __name__ == '__main__':
    run()



#!/usr/bin/env python
import os
import sys
from setuptools import setup
from textwrap import dedent

NAME = "docker-zabbix-sender"
GITHUB_ORG_URL = "https://github.com/dockermeetupsinbordeaux"
ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

exec(open('docker_zabbix_sender/version.py').read())

setup(
    name=NAME,
    version=version,
    author="Tristan Carel",
    author_email="tristan.carel@gmail.com",
    url= GITHUB_ORG_URL + '/' + NAME,
    download_url="{0}/{1}/tarball/v{2}".format(GITHUB_ORG_URL, NAME, version),
    description="Push Docker containers statistics to Zabbix efficiently",
    long_description=dedent("""
        Rationale
        ---------
        Docker Zabbix Sender delivers a daemon script that push to Zabbix statistics about Docker containers.

        It leverages 2 interesting components:

        - Zabbix maintains a tool titled ``zabbix-sender``.
          It is meant to push `Zabbix trapper items`_ efficiently.

        - Docker 1.5.0 comes with Docker Remote API version 17, providing a new `stats endpoint`_.
          It allows the client to subscribe to a live feed delivering a container statistics.

        The daemon script stands in the middle of those 2 components.
        It collects Docker containers statistics and transforms them in Zabbix trapper events.

        Published metrics
        -----------------
        The daemon script already emits a collection of metrics: CPU, memory, and network usage.
        The module's API provide mechanism to easily extend the list of published events.

        Documentation
        -------------
        The stable documentation is available on ReadTheDocs_

        .. _Zabbix trapper items: https://www.zabbix.com/documentation/2.4/manual/config/items/itemtypes/trapper
        .. _stats endpoint: https://docs.docker.com/reference/api/docker_remote_api_v1.17/#get-container-stats-based-on-resource-usage
        .. _ReadTheDocs: http://docker-zabbix-sender.readthedocs.org/en/stable/
    """),
    keywords="docker zabbix monitoring",
    packages=['docker_zabbix_sender'],
    install_requires=[
        'docker-py >= 1.0.0',
    ],
    zip_safe=False,
    license="Apache license version 2.0",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Utilities',
        'License :: OSI Approved :: Apache Software License',
    ],
    entry_points = """
        [console_scripts]
        docker-zabbix-sender = docker_zabbix_sender.zabbix_sender:run
        [docker_zabbix_sender.metrics]
        containers-count = docker_zabbix_sender.stats:containers_count
        cpu_count = docker_zabbix_sender.stats:cpu_count
    """
)
